from sparknlp_jsl.common import *
from sparknlp.annotator.classifier_dl import RoBertaForQuestionAnswering


class ZeroShotNerModel(RoBertaForQuestionAnswering, HasEngine):
    """
    Zero shot named entity recognition based on RoBertaForQuestionAnswering.

    ==========================================  ======================
    Input Annotation types                      Output Annotation type
    ==========================================  ======================
    ``DOCUMENT, TOKEN``                         ``NAMED_ENTITY``
    ==========================================  ======================

    Parameters
    ----------
    entityDefinitions
        A dictionary with definitions of named entities. The keys of dictionary are the entity labels and the
        values are lists of questions. For example:
              {
                "CITY": ["Which city?", "Which town?"],
                "NAME": ["What is her name?", "What is his name?"]}

    predictionThreshold
        Minimal confidence score to encode an entity (Default: 0.01f)

    ignoreEntities
        A list of entity labels which are discarded from the output.

    Examples
    --------
    --------
    >>> document_assembler = DocumentAssembler() \
    ...     .setInputCol("text") \
    ...     .setOutputCol("document")
    >>> sentence_detector = annotators.SentenceDetector()\
    ...     .setInputCols(["document"])\
    ...     .setOutputCol("sentence")
    >>> tokenizer = annotators.Tokenizer()\
    ...     .setInputCols(["sentence"])\
    ...     .setOutputCol("token")
    >>> zero_shot_ner = ZeroShotNerModel()\
    ...     .load("/models/sparknlp/zero_shot_ner")\
    ...     .setEntityDefinitions(
    ...         {
    ...             "NAME": ["What is his name?", "What is my name?", "What is her name?"],
    ...             "CITY": ["Which city?", "Which is the city?"]
    ...         })\
    ...     .setInputCols(["sentence", "token"])\
    ...     .setOutputCol("zero_shot_ner")\
    >>> data = spark.createDataFrame(
    ...         [["My name is Clara, I live in New York and Hellen lives in Paris."]]
    ...     ).toDF("text")
    >>> Pipeline() \
    ...     .setStages([document_assembler, sentence_detector, tokenizer, zero_shot_ner]) \
    ...     .fit(data) \
    ...     .transform(data) \
    ...     .selectExpr("document", "explode(zero_shot_ner) AS entity")\
    ...     .select(
    ...         "document.result",
    ...         "entity.result",
    ...         "entity.metadata.word",
    ...         "entity.metadata.confidence",
    ...         "entity.metadata.question")\
    ...     .show(truncate=False)
        +-----------------------------------------------------------------+------+------+----------+------------------+
        |result                                                           |result|word  |confidence|question          |
        +-----------------------------------------------------------------+------+------+----------+------------------+
        |[My name is Clara, I live in New York and Hellen lives in Paris.]|B-CITY|Paris |0.5328949 |Which is the city?|
        |[My name is Clara, I live in New York and Hellen lives in Paris.]|B-NAME|Clara |0.9360068 |What is my name?  |
        |[My name is Clara, I live in New York and Hellen lives in Paris.]|B-CITY|New   |0.83294415|Which city?       |
        |[My name is Clara, I live in New York and Hellen lives in Paris.]|I-CITY|York  |0.83294415|Which city?       |
        |[My name is Clara, I live in New York and Hellen lives in Paris.]|B-NAME|Hellen|0.45366877|What is her name? |
        +-----------------------------------------------------------------+------+------+----------+------------------+
    """
    inputAnnotatorTypes = [AnnotatorType.DOCUMENT, AnnotatorType.TOKEN]
    outputAnnotatorType = AnnotatorType.NAMED_ENTITY

    name = "ZeroShotNerModel"

    predictionThreshold = Param(Params._dummy(),
                                "predictionThreshold",
                                "Minimal confidence score to encode an entity (default is 0.1)",
                                TypeConverters.toFloat)

    ignoreEntities = Param(Params._dummy(),
                           "ignoreEntities",
                           "List of entities to ignore",
                           TypeConverters.toListString)

    def setPredictionThreshold(self, threshold):
        """Sets the minimal confidence score to encode an entity

        Parameters
        ----------
        threshold : float
           minimal confidence score to encode an entity (default is 0.1)
        """
        return self._set(predictionThreshold=threshold)

    def setEntityDefinitions(self, definitions):
        """Set entity definitions

        Parameters
        ----------
        definitions : dict[str, list[str]]

        """
        self._call_java("setEntityDefinitions", definitions)

        return self

    def getClasses(self):
        """
        Returns the list of entities which are recognized
        """
        return self._call_java("getEntities")

    @keyword_only
    def __init__(self, classname="com.johnsnowlabs.nlp.annotators.ner.ZeroShotNerModel", java_model=None):
        super(ZeroShotNerModel, self).__init__(
            classname=classname,
            java_model=java_model
        )
        self._setDefault(
            predictionThreshold=0.01,
            ignoreEntities=[]
        )

    @staticmethod
    def pretrained(name="zero_shot_ner_roberta", lang="en", remote_loc="clinical/models"):
        from sparknlp_jsl.pretrained import InternalResourceDownloader
        return InternalResourceDownloader.downloadModel(ZeroShotNerModel, name, lang, remote_loc,
                                                        j_dwn='InternalsPythonResourceDownloader')

    @staticmethod
    def load(path):
        from sparknlp_jsl.internal import _RobertaQAToZeroShotNerLoader
        jModel = _RobertaQAToZeroShotNerLoader(path)._java_obj
        return ZeroShotNerModel(java_model=jModel)